/*
  This file is part of the Changedfiles system. 
  $Id: changedfiles.c,v 1.13 2003/02/08 04:16:56 rocky Exp $

  This product includes software developed by Konrad Rieck, modified
  by Rocky Bernstein.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or (at
  your option) any later version.

  This program is distributed in the hope that it will be useful, but
  WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
  USA
******************************************************************************/
/*
  The main routines (initialization and finalization) for the
  Changefiles module. 
*/

#define _KERNEL

#include "headers.h"
#include "syscalls.h"
#include "comm.h"

/*
 * Runtime configuration
 */
chf_config_t config = {
    /* debug, ... */
    ON, 
};

/*
 * This is the loadable module wrapper.
 */
extern struct mod_ops mod_miscops;

/*
 * Structure of the system-entry table.
 */
#ifdef _64BIT
extern struct sysent sysent32[];
#else
extern struct sysent sysent[];
#endif

/*
 * Pointer to the proc filesystem vnode operations. 
 */
static struct vnodeops *procfs_vnodeops;

/* 
 * Structure of the loaded modules
 */
extern struct modctl modules;

/*
 * Module linkage information for the kernel.
 */
static struct modlmisc modlmisc = {
    &mod_miscops,
    "Changedfiles v" VERSION
};

struct modlinkage modlinkage = {
    MODREV_1,
    { (void *) &modlmisc, NULL }
};

/*
 * Syscall redirection vector table. 
 * NOTE: The order here has to match the enumeration syscall_op
 * in changedfiles.h
 */
chf_syscall_t syscalls[] = {

  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*open*/
  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*open64*/
  {0, (int64_t(*)())chf_rmdir,
   SE_NOUNLOAD | SE_LOADED, SYS_rmdir,   DFLT_ARG, 0},      /*rmdir*/
  {0, (int64_t(*)())chf_mkdir, 
   SE_NOUNLOAD | SE_LOADED, SYS_mkdir,   DFLT_ARG, 0},      /*mkdir*/
  {0, (int64_t(*)())chf_symlink, 
   SE_NOUNLOAD | SE_LOADED, SYS_symlink, DFLT_ARG, 0},      /*symlink*/
  {0, (int64_t(*)())chf_link, 
   SE_NOUNLOAD | SE_LOADED, SYS_link,    DFLT_ARG, 0},      /*link*/
#if 0
  {0, (int64_t(*)())chf_unlink, 
   SE_NOUNLOAD | SE_LOADED, SYS_unlink,  DFLT_ARG, 0},      /*unlink*/
#else 
  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*unlink notouch*/
#endif
  {0, (int64_t(*)())chf_creat, 
   SE_NOUNLOAD | SE_LOADED, SYS_creat,   DFLT_ARG, 0},      /*creat*/
  {0, (int64_t(*)())chf_creat64, 
   SE_NOUNLOAD | SE_LOADED, SYS_creat64, DFLT_ARG, 0},      /*creat64*/
  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*close*/
  {0, (int64_t(*)())chf_rename, 
   SE_NOUNLOAD | SE_LOADED, SYS_rename, DFLT_ARG, 0},       /*rename*/
  {0, (int64_t(*)())chf_comm, SE_NOUNLOAD | SE_LOADED, SYS_chfcomm, 2, 0},
  {0, (int64_t(*)())chf_read, SE_NOUNLOAD | SE_LOADED, SYS_chfread, 1, 0},
  {0, 0, 0, 0, 0, 0}
};

#ifdef _64BIT
/*
  Syscall redirection vector table. 
  NOTE: The order here has to match the enumeration syscall_id
  in changedfiles.h
 */
chf_syscall_t syscalls32[] = {

  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*open*/
  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*open64*/
  {0, (int64_t(*)())chf_rmdir,
   SE_NOUNLOAD | SE_LOADED, SYS_rmdir,   DFLT_ARG, 0},      /*rmdir*/
  {0, (int64_t(*)())chf_mkdir, 
   SE_NOUNLOAD | SE_LOADED, SYS_mkdir,   DFLT_ARG, 0},      /*mkdir*/
  {0, (int64_t(*)())chf_symlink, 
   SE_NOUNLOAD | SE_LOADED, SYS_symlink, DFLT_ARG, 0},      /*symlink*/
  {0, (int64_t(*)())chf_link, 
   SE_NOUNLOAD | SE_LOADED, SYS_link,    DFLT_ARG, 0},      /*link*/
  {0, (int64_t(*)())chf_unlink, 
   SE_NOUNLOAD | SE_LOADED, SYS_unlink,                     /*unlink*/
   DFLT_ARG, 0},
  {0, (int64_t(*)())chf_creat32, 
   SE_NOUNLOAD | SE_LOADED, SYS_creat,   DFLT_ARG, 0},      /*creat*/
  {0, (int64_t(*)())chf_creat32_64, 
   SE_NOUNLOAD | SE_LOADED, SYS_creat64, DFLT_ARG, 0},      /*creat64*/
  {0, 0, SE_NOUNLOAD | SE_LOADED, SYS_UNUSED, DFLT_ARG, 0}, /*close*/
  {0, (int64_t(*)())chf_rename, 
   SE_NOUNLOAD | SE_LOADED, SYS_rename, DFLT_ARG, 0},       /*rename*/
  {0, (int64_t(*)())chf_comm, SE_NOUNLOAD | SE_LOADED, SYS_chfcomm, 2, 0},
  {0, (int64_t(*)())chf_read, SE_NOUNLOAD | SE_LOADED, SYS_chfread, 1, 0},
  {0, 0, 0, 0, 0, 0}
};

#endif				/* _64BIT */

/*
 * Mutexes, Locks, condition variables.
 */
kmutex_t   readdir_lock;
kmutex_t   refcnt_lock;
kmutex_t   queue_lock;       /* Lock for event queue update */
kmutex_t   read_ready_lock;  /* To make syscall synchronous with data */
kmutex_t   write_ready_lock; /* To not write before last read acknowleged */
kcondvar_t read_ready_cv;    /* Condition variable we signal on read */
kcondvar_t write_ready_cv;   /* Condition variable we signal on write */

krwlock_t  config_rwlock;

/*
 * Restore the old syscall vectors saved in the syscall struct.
 */
void 
restore_old_syscalls()
{
    int i;

    for (i = 0; syscalls[i].index; i++) {
	if (syscalls[i].index != SYS_UNUSED) {
	    dcmn_err((CE_CONT, "Restoring syscall %d.\n",
		      syscalls[i].index));
#ifdef _64BIT
	    sysent32[syscalls32[i].index].sy_callc = syscalls32[i].sc;
	    sysent32[syscalls32[i].index].sy_flags = syscalls32[i].flags;

	    if (syscalls32[i].narg != DFLT_ARG)
		sysent32[syscalls32[i].index].sy_narg = syscalls32[i].onarg;
#endif

	    sysent[syscalls[i].index].sy_callc = syscalls[i].sc;
	    sysent[syscalls[i].index].sy_flags = syscalls[i].flags;

	    if (syscalls[i].narg != DFLT_ARG)
		sysent[syscalls[i].index].sy_narg = syscalls[i].onarg;
	}
    }

}

/*
 * Store the old syscall vectors in the syscall struct.
 */
void 
store_old_syscalls()
{
    int i;

    for (i = 0; syscalls[i].index; i++) {
	if (syscalls[i].index != SYS_UNUSED) {
	    dcmn_err((CE_CONT, "Storing syscall %d.\n", syscalls[i].index));

#ifdef _64BIT
	    syscalls32[i].sc =
		(int64_t(*)())sysent32[syscalls32[i].index].sy_callc;
	    syscalls32[i].flags = sysent32[syscalls32[i].index].sy_flags;
	    syscalls32[i].onarg = sysent32[syscalls32[i].index].sy_narg;
#endif

	    syscalls[i].sc =
		(int64_t(*)())sysent[syscalls[i].index].sy_callc;
	    syscalls[i].flags = sysent[syscalls[i].index].sy_flags;
	    syscalls[i].onarg = sysent[syscalls[i].index].sy_narg;

	}
    }
}

/*
 * Set the new syscall vectors specified in the syscall struct.
 */
void 
set_new_syscalls()
{
    int i;

    for (i = 0; syscalls[i].index; i++) {
	if (syscalls[i].index != SYS_UNUSED) {
	    dcmn_err((CE_CONT, "Setting new syscall %d.\n",
		      syscalls[i].index));
#ifdef _64BIT

	    if (syscalls32[i].sc == syscalls[i].sc &&
		sysent32[syscalls32[i].index].sy_callc !=
		sysent[syscalls[i].index].sy_callc)
		cmn_err(CE_PANIC, "32 and 64 bit syscalls are not "
			"identical by default, but changefiles claims they "
			"are!");

	    sysent32[syscalls32[i].index].sy_callc = syscalls32[i].nc;
	    sysent32[syscalls32[i].index].sy_flags = syscalls32[i].flags;

	    if (syscalls32[i].narg != DFLT_ARG)
		sysent32[syscalls32[i].index].sy_narg = syscalls32[i].narg;
#endif
	    sysent[syscalls[i].index].sy_callc = syscalls[i].nc;
	    sysent[syscalls[i].index].sy_flags = syscalls[i].flags;

	    if (syscalls[i].narg != DFLT_ARG)
		sysent[syscalls[i].index].sy_narg = syscalls[i].narg;
	}
    }
}

/*
 * Init and load the module.
 */
int
_init()
{
    int i, cri;
    struct modctl *this;

    procfs_vnodeops = (struct vnodeops *) modgetsymvalue("prvnodeops", 0);

    if (!procfs_vnodeops) {
	cmn_err(CE_WARN, "Symbol prvnodeops not found, procfs module not "
		" loaded.");
	return set_errno(ENOENT);
    }

    if ((i = mod_install(&modlinkage)) != 0) {
	cmn_err(CE_WARN, "Could not load module.");
	return i;
    } else {
	cmn_err(CE_CONT, "Changedfiles v%s successfully loaded.\n", VERSION);
    }

    /*
     * Initialize mutexes and locks
     */
    dcmn_err((CE_CONT, "Initializing locks.\n"));
    mutex_init(&refcnt_lock, NULL, MUTEX_DEFAULT, NULL);
    mutex_init(&readdir_lock, NULL, MUTEX_DEFAULT, NULL);
    mutex_init(&read_ready_lock, NULL, MUTEX_DEFAULT, NULL);
    mutex_init(&write_ready_lock, NULL, MUTEX_DEFAULT, NULL);
    mutex_init(&queue_lock,  NULL, RW_DRIVER, NULL);
    cv_init(&read_ready_cv, NULL, CV_DRIVER, NULL);
    cv_init(&write_ready_cv, NULL, CV_DRIVER, NULL);

    rw_init(&config_rwlock, NULL, RW_DRIVER, NULL);

    /*
     * Module is loaded, find it in the modules list
     */
    this = &modules;
    while (this->mod_linkage != &modlinkage) {
	this = this->mod_next;
	ASSERT(this != &modules);
    }

    /*
     * Initialize reference counter
     */
    init_refcnt();

    dcmn_err((CE_CONT, "Entering critical area.\n"));
    cri = ddi_enter_critical();

    /*
     * Redirect syscalls. 
     */
    store_old_syscalls();
    set_new_syscalls();

    ddi_exit_critical(cri);
    dcmn_err((CE_CONT, "Leaving critical area.\n"));

    return i;
}

int
_info(struct modinfo *modinfop)
{
    return (mod_info(&modlinkage, modinfop));
}

/* 
 * "fini" and unload the module.
 */
int
_fini()
{
    int i, cri;

    if (!check_refcnt())
	return set_errno(EBUSY);

    dcmn_err((CE_CONT, "Entering critical area.\n"));
    cri = ddi_enter_critical();

    /*
     * Restore old syscall vectors. 
     */
    restore_old_syscalls();
    ddi_exit_critical(cri);
    chf_queue_destroy();
    dcmn_err((CE_CONT, "Leaving critical area.\n"));

    /*
     * Destroy mutexes.
     */
    dcmn_err((CE_CONT, "Destroying locks.\n"));
    rw_destroy(&config_rwlock);
    mutex_destroy(&queue_lock);
    mutex_destroy(&read_ready_lock);
    mutex_destroy(&write_ready_lock);
    mutex_destroy(&readdir_lock);
    mutex_destroy(&refcnt_lock);
    cv_destroy(&read_ready_cv);
    cv_destroy(&write_ready_cv);

    if ((i = mod_remove(&modlinkage)) != 0)
	cmn_err(CE_WARN, "Could not unload module.");
    else
	cmn_err(CE_CONT, "Changedfiles v%s successfully unloaded.\n", VERSION);

    return i;
}
