/*****************************************************************************
     This file is part of the Changedfiles system. 

     Copright (C)2001,2002  Michael L. Welles 
     Modified for Solaris by Rocky Bernstein

     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
******************************************************************************/

/* $Id: dev.c,v 1.1 2002/12/29 21:07:46 rocky Exp $ */ 

#define _KERNEL

#include <limits.h>       /* PATH_MAX */
#include <sys/copyops.h>  /* copyinstr() */
#include "headers.h"
#include "libc.h"

/* buffer for character device */

#define BUF_SIZE        512
char *buffer[BUF_SIZE];
char *read_buf; 

int queue_head = 0;
int queue_tail = 0;

/* device info for the changedfiles device */ 
static int changedfiles_major = 40;

extern krwlock_t queue_rwlock;


int 
changedfiles_get_from_queue(char * c, int size) 
{
  rw_enter(&queue_rwlock, RW_READER);
  *c='\0'; 
  
  if (queue_head != queue_tail) {
    chf_strncpy(c,buffer[queue_head],size);
    kmem_free(buffer[queue_head], size);
  
    queue_head++;
    if(queue_head == BUF_SIZE) queue_head=0;
  }
  
  rw_exit(&queue_rwlock); 
  return 0;
}

int 
changedfiles_put_to_queue(char * c)
{
  int len; 
  dcmn_err((CE_NOTE, "put_to_queue: %s", c)); 
  rw_enter(&queue_rwlock, RW_READER);
  if( (queue_tail + 1) == queue_head
      || (
	  ((queue_tail + 1) == BUF_SIZE)
	  && ( queue_head == 0 ) ))
    { 
      queue_head++;
      if ( queue_head == BUF_SIZE ) 
	{ 
	  queue_head = 0; 
	}
    }
  len = chf_strlen(c); 
  len++; 
  buffer[queue_tail] = kmem_zalloc(len, KM_SLEEP); 
  if ( ! buffer[queue_tail] ) 
    { 
      rw_exit(&queue_rwlock); 
      return -ENOMEM; 
    }
  chf_strncpy(buffer[queue_tail], c, len); 
  queue_tail++;
  
  if ( queue_tail == BUF_SIZE ) 
    queue_tail=0;
  rw_exit(&queue_rwlock);
  return 1;
}

int 
changedfiles_open_dev( struct inode *in, struct file * fi )
{ 
  /* FIXME: MOD_INC_USE_COUNT; */
  dcmn_err((CE_NOTE, "changedfiles_open_dev()")); 
  return 0; 
}

int 
changedfiles_close_dev( struct inode *in, struct file * fi )
{
  dcmn_err((CE_NOTE, "changedfiles_close_dev()")); 
  /* FIXME: MOD_DEC_USE_COUNT;  */
  return 0; 
}


ssize_t 
changedfiles_read_dev( struct file *filep, char *buf, size_t count, 
		       off_t *f_pos )
{ 
  
  int len; 
  int mycount; 
  
  if ( count > PATH_MAX ) 
     { 
       mycount = PATH_MAX; 
     }
  else 
    { 
      mycount = count; 
    }
  
  read_buf = kmem_zalloc( mycount, KM_SLEEP); 
  
  if ( !read_buf ) { 
    return -ENOMEM; 
  }
  chf_memset(read_buf, 0, mycount); 
  changedfiles_get_from_queue(read_buf, mycount);
  len = chf_strlen(read_buf); 
  if ( len == 0 ) 
    { 
      kmem_free(read_buf, 1); 
      return 0; 
    }
  
  dcmn_err((CE_NOTE, "got '%s' from queue", read_buf));
  /* include the \0 */ 
  len++; 
  if (copyoutstr(read_buf, buf, mycount, 0) )
    { 
      dcmn_err((CE_NOTE, "copyoutstr failed")); 
      return -EFAULT; 
    }
  dcmn_err((CE_NOTE, "changedfiles_read_dev() returning '%s'", 
	    read_buf));
  return len; 
}

static struct file_operations changedfiles_fop = { 
  read: changedfiles_read_dev, 
  open: changedfiles_open_dev, 
  release: changedfiles_close_dev, 
}; 


  




