/*****************************************************************************
 This file is part of the Changedfiles system. 
 $Id: libc.c,v 1.4 2003/02/03 04:43:00 rocky Exp $

 This product includes software developed by Konrad Rieck, modified
 by Rocky Bernstein.

 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or (at
 your option) any later version.

 This program is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 USA
******************************************************************************/
/*
  This file contains things that would normally be in libc, but since
  we're running in the kernel are not available.
*/

#define _KERNEL

#include "headers.h"

#include <limits.h>
#include <unistd.h>
/* This is to prevent a conflicting definition of ffs in string.h */
#define _POSIX_C_SOURCE 1
#include <string.h>

/*!
  Turn unsigned int i into an ASCII string stored into str
  and return the length of that. Length 0 means error.
*/
unsigned int
chf_uitostr ( char * str, unsigned int i) 
{ 
  char c[30]; 
  int j=i; 
  int len=0;
  if (j==0) {
    str[0]='0';
    str[1]='\0';
    return 1;
  } 
  while (j != 0) {
    c[len++] = '0' + j % 10;
    j /= 10;
  }
  for (j=1; j<=len; j++) {
    *str++=c[len-j];
  }
  *str='\0';
  return len; 
}

/*!
  Classical memset implementation used for compilers that play dirty macro
  tricks with me and my headers and rely on memset(). :(
*/
void *
chf_memset(void *sp1, int c, size_t n)
{
    if (n != 0) {
	unsigned char *sp = sp1;
	do {
	    *sp++ = (unsigned char) c;
	} while (--n != 0);
    }
    return (sp1);
}

int 
chf_strlen ( const char * mstr) 
{ 
  char * c; 
  int i=0; 
  c = (char *)mstr; 
  while ( *c++ != '\0' ) i++;
  return i; 
}

/* Taken from glibc */
char *
chf_strncat (char *s1, const char *s2, size_t n)
{
  char c;
  char *s = s1;

  /* Find the end of S1.  */
  do
    c = *s1++;
  while (c != '\0');

  /* Make S1 point before next character, so we can increment
     it while memory is read (wins on pipelined cpus).  */
  s1 -= 2;

  if (n >= 4)
    {
      size_t n4 = n >> 2;
      do
	{
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    return s;
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    return s;
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    return s;
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    return s;
	} while (--n4 > 0);
      n &= 3;
    }

  while (n > 0)
    {
      c = *s2++;
      *++s1 = c;
      if (c == '\0')
	return s;
      n--;
    }

  if (c != '\0')
    *++s1 = '\0';

  return s;
}

/* Taken from glibc */
char *
chf_strncpy (char *s1, const char *s2, size_t n)
{
  char c;
  char *s = s1;

  --s1;

  if (n >= 4)
    {
      size_t n4 = n >> 2;

      for (;;)
	{
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    break;
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    break;
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    break;
	  c = *s2++;
	  *++s1 = c;
	  if (c == '\0')
	    break;
	  if (--n4 == 0)
	    goto last_chars;
	}
      n = n - (s1 - s) - 1;
      if (n == 0)
	return s;
      goto zero_fill;
    }

 last_chars:
  n &= 3;
  if (n == 0)
    return s;

  do
    {
      c = *s2++;
      *++s1 = c;
      if (--n == 0)
	return s;
    }
  while (c != '\0');

 zero_fill:
  do
    *++s1 = '\0';
  while (--n > 0);

  return s;
}
